﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace DeviceWatcher
{

    public enum CMD : uint
    {
        APPEND              = 4294967070,
        BGCOLOR             = 4294967049,
        BITMAP_TRANSFORM    = 4294967073,
        BUTTON              = 4294967053,
        CALIBRATE           = 4294967061,
        CLOCK = 4294967060,
        COLDSTART = 4294967090,
        CRC = 4294967043,
        DIAL = 4294967085,
        DLSTART = 4294967040,
        EXECUTE = 4294967047,
        FGCOLOR = 4294967050,
        GAUGE = 4294967059,
        GETMATRIX = 4294967091,
        GETPOINT = 4294967048,
        GETPROPS = 4294967077,
        GETPTR = 4294967075,
        GRADCOLOR = 4294967092,
        GRADIENT = 4294967051,
        HAMMERAUX = 4294967044,
        IDCT = 4294967046,
        INFLATE = 4294967074,
        INTERRUPT = 4294967042,
        KEYS = 4294967054,
        LOADIDENTITY = 4294967078,
        LOADIMAGE = 4294967076,
        LOGO = 4294967089,
        MARCH = 4294967045,
        MEMCPY = 4294967069,
        MEMCRC = 4294967064,
        MEMSET = 4294967067,
        MEMWRITE = 4294967066,
        MEMZERO = 4294967068,
        NUMBER = 4294967086,
        PROGRESS = 4294967055,
        REGREAD = 4294967065,
        ROTATE = 4294967081,
        SCALE = 4294967080,
        SCREENSAVER = 4294967087,
        SCROLLBAR = 4294967057,
        SETFONT = 4294967083,
        SETMATRIX = 4294967082,
        SKETCH = 4294967088,
        SLIDER = 4294967056,
        SNAPSHOT = 4294967071,
        SPINNER = 4294967062,
        STOP = 4294967063,
        SWAP = 4294967041,
        TEXT = 4294967052,
        TOGGLE = 4294967058,
        TOUCH_TRANSFORM = 4294967072,
        TRACK = 4294967084,
        TRANSLATE = 4294967079
    }
 
    class CoPro :  EveDL
    {
         private const int FT_CMD_SIZE = 4;


        //public void Flush()
        //{


        //}
        public class Button : FTEveCmd
        {
            CMD    c;
            uint   size;
            short  x;
            short  y;
            short  w;
            short  h;
            short  font;
            ushort options;
            string text;


            /// <summary>
            /// Button
            /// </summary>
            /// <param name="inX">x</param>
            /// <param name="inY">y</param>
            /// <param name="inW">w</param>
            /// <param name="inH">h</param>
            /// <param name="inFont">font</param>
            /// <param name="inOptions">options</param>
            /// <param name="inText">text</param>
            public Button(short inX, short inY, short inW, short inH, short inFont, ushort inOptions, string inText)
            {
                c       =  CMD.BUTTON;
                x       =  inX;
                y       =  inY;
                w       =  inW;
                h       =  inH;
                font    =  inFont;
                options =  inOptions;
                text    =  inText;
                uint z       = 16 + (uint)text.Length + 1;
                size = (z + 3) - ((z + 3) % 4);  // extends to 32 bit / 4 byte boundaries...

            }
            public CMD cmd()
            {
                return c;
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                byte[] tText = Encoding.UTF8.GetBytes(text);
                byte[] buff = new byte[size];

                int i = 0;
                buff[i++] = (byte)((uint)c & 0x000000FF);              // Command
                buff[i++] = (byte)(((uint)c & 0x0000FF00) >> 8);
                buff[i++] = (byte)(((uint)c & 0x00FF0000) >> 16);
                buff[i++] = (byte)(((uint)c & 0xFF000000) >> 24);

                buff[i++] = (byte)((uint) x & 0x000000FF);              // x coord
                buff[i++] = (byte)(((uint)x & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) y & 0x000000FF);              // y coord
                buff[i++] = (byte)(((uint)y & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) w & 0x000000FF);              // w
                buff[i++] = (byte)(((uint)w & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) h & 0x000000FF);              // h
                buff[i++] = (byte)(((uint)h & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) font & 0x000000FF);           //font
                buff[i++] = (byte)(((uint)font & 0x0000FF00) >> 8);


                buff[i++] = (byte)((uint) options & 0x000000FF);        //options
                buff[i++] = (byte)(((uint)options & 0x0000FF00) >> 8);

                tText.CopyTo(buff, i);                                  // text
                i += text.Length;
                do
                {
                    buff[i++] = 0;                                        // null
                } while (i < buff.Length);
                return buff;
            }
        }

        //
        /// <summary>
        ///  Co processor instructions
        /// /
        /// </summary>
        /// 
        public class Flush : FTEveCmd
        {

            override public uint Size()
            {
                return 0;
            }

            override public byte[] ToBuffer()
            {
              return new byte[]{ };  
            }
        }


        //
        /// <summary>
        ///  Co processor instructions
        /// /
        /// </summary>
        public class Getpoint : FTEveCmd
        {
            public CMD Cmd()
            {
                throw new NotImplementedException();
            }

            override public uint Size()
            {
                throw new NotImplementedException();
            }

            override public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }

        public class Dlstart : FTEveCmd
        {
            const CMD c = CMD.DLSTART;
            const uint size = 4;
            public Dlstart()
            {

            }
            public CMD Cmd()
            {
                return c;
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                return CMDToBuffer(c);
            }
        }

        public class Gauge : FTEveCmd
        {
            CMD c;
            uint size;
            short x;
            short y;
            short r;
            ushort options;
            ushort major;
            ushort minor;
            ushort val;
            ushort range;            

            /// <summary>
            /// Gauge
            /// </summary>
            /// <param name="inX">x</param>
            /// <param name="inY">y</param>
            /// <param name="inR">r</param>
            /// <param name="inOptions">options</param>
            /// <param name="inMajor">major</param>
            /// <param name="inMinor">minor</param>
            /// <param name="inVal">value</param>
            /// <param name="inRange">range</param>
            public Gauge(short inX, short inY, short inR, ushort inOptions, ushort inMajor, ushort inMinor, ushort inVal, ushort inRange )
            {
                c = CMD.GAUGE;
                x = inX;
                y = inY;
                r = inR;
                options = inOptions;
                major   = inMajor;
                minor   = inMinor;
                val     = inVal;
                range   = inRange;
                                           
                
                size = 20;  // extends to 32 bit / 4 byte boundaries...

            }
            public CMD cmd()
            {
                return c;
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];

                int i = 0;
                buff[i++] = (byte)((uint)c & 0x000000FF);               // Command
                buff[i++] = (byte)(((uint)c & 0x0000FF00) >> 8);
                buff[i++] = (byte)(((uint)c & 0x00FF0000) >> 16);
                buff[i++] = (byte)(((uint)c & 0xFF000000) >> 24);

                buff[i++] = (byte)((uint)x & 0x000000FF);               // x coord
                buff[i++] = (byte)(((uint)x & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint)y & 0x000000FF);               // y coord
                buff[i++] = (byte)(((uint)y & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) r & 0x000000FF);              // r
                buff[i++] = (byte)(((uint)r & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) options & 0x000000FF);        // options
                buff[i++] = (byte)(((uint)options & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) major & 0x000000FF);          // major
                buff[i++] = (byte)(((uint)major & 0x0000FF00) >> 8);


                buff[i++] = (byte)((uint) minor & 0x000000FF);          //minor
                buff[i++] = (byte)(((uint)minor & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) val & 0x000000FF);        // value
                buff[i++] = (byte)(((uint)val & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) range & 0x000000FF);        //range
                buff[i++] = (byte)(((uint)range & 0x0000FF00) >> 8);               
                return buff;
            }
        }

        public class Swap : FTEveCmd
        {
            const CMD c = CMD.SWAP;
            const uint size = 4;

            public Swap()
            {
                            }
            public CMD Cmd()
            {
                return c;
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                return CMDToBuffer(c);
            }
        }
        public class Text : FTEveCmd
        {
            CMD c;
            uint x;
            uint y;
            uint font;
            byte options;
            string text;
            uint state;
            int size;


            public Text (uint inx, uint  iny, uint infont, byte inoptions, string intext)
            {
                int z;

                x = inx;
                y = iny;
                font = infont;
                options = inoptions;
                text = intext;
                                
                z = 12 + intext.Length + 1;
                size = (z + 3) - ((z + 3) % 4);  // extends to 32 bit / 4 byte boundaries...
                c = CMD.TEXT;
                
            }

            override public uint Size()
            {
                return (uint)size;
            }

            override public byte[] ToBuffer()
            {
                byte[] tText = Encoding.UTF8.GetBytes(text);
                byte[] buff = new byte[size];

                int i = 0;
                buff[i++] = (byte)((uint)c & 0x000000FF);              // Command
                buff[i++] = (byte)(((uint)c & 0x0000FF00) >> 8);
                buff[i++] = (byte)(((uint)c & 0x00FF0000) >> 16);
                buff[i++] = (byte)(((uint)c & 0xFF000000) >> 24);

                buff[i++] = (byte)((uint)x & 0x000000FF);              // x coord
                buff[i++] = (byte)(((uint)x & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint)y & 0x000000FF);              // y coord
                buff[i++] = (byte)(((uint)y & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint)font & 0x000000FF);           //font
                buff[i++] = (byte)(((uint)font & 0x0000FF00) >> 8);


                buff[i++] = (byte)((uint)options & 0x000000FF);        //options
                buff[i++] = (byte)(((uint)options & 0x0000FF00) >> 8);

                tText.CopyTo(buff, i);                                  // text             
                i += text.Length;
                do
                {
                    buff[i++] = 0;                                        // null
                } while (i < buff.Length);
                return buff;                
            }
            public CMD Cmd()
            {
                return c;
            }
        }

        public class Gradient : FTEveCmd
        {
            private CMD      c;
            private short    x0;
            private short    y0;
            private uint     rgb0;
            private short    x1;
            private short    y1;
            private uint     rgb1;
            private uint size = 20;
            public Gradient(short inX0, short inY0, uint inRgb0,short inX1, short inY1, uint inRgb1)
            {
                c = CMD.GRADIENT;
                x0      = inX0;
                y0      = inY0;
                rgb0    = inRgb0;
                x1      = inX1;
                y1      = inY1;
                rgb1    = inRgb1;
            }
            public CMD cmd()
            {
                return c;
            }

            override public uint Size()
            {
                return size;
            }

            override public byte[] ToBuffer()
            {
                byte[] buff = new byte[size];

                int i = 0;
                buff[i++] = (byte)((uint) c & 0x000000FF);                  // Command
                buff[i++] = (byte)(((uint)c & 0x0000FF00) >> 8);
                buff[i++] = (byte)(((uint)c & 0x00FF0000) >> 16);
                buff[i++] = (byte)(((uint)c & 0xFF000000) >> 24);
                             
                buff[i++] = (byte)((uint) x0 & 0x000000FF);                 // x0
                buff[i++] = (byte)(((uint)x0 & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) y0 & 0x000000FF);                 // y0
                buff[i++] = (byte)(((uint)y0 & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) rgb0 & 0x000000FF);               // rgb0
                buff[i++] = (byte)(((uint)rgb0 & 0x0000FF00) >> 8);
                buff[i++] = (byte)(((uint)rgb0 & 0x00FF0000) >> 16);
                buff[i++] = (byte)(((uint)rgb0 & 0xFF000000) >> 24);

                buff[i++] = (byte)((uint) x1 & 0x000000FF);                 // x1
                buff[i++] = (byte)(((uint)x1 & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) y1 & 0x000000FF);                 // y1
                buff[i++] = (byte)(((uint)y1 & 0x0000FF00) >> 8);

                buff[i++] = (byte)((uint) rgb1 & 0x000000FF);               // rgb1
                buff[i++] = (byte)(((uint)rgb1 & 0x0000FF00) >> 8);
                buff[i++] = (byte)(((uint)rgb1 & 0x00FF0000) >> 16);
                buff[i++] = (byte)(((uint)rgb1 & 0xFF000000) >> 24);

                return buff;
            }
        }



        public class Toggle : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            override public uint Size()
            {
                throw new NotImplementedException();
            }

            override public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }

#if laters

        class Append : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Bgcolor : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Bitmap_transfor : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
      
        class Calibrate : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Clock : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Coldstart : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Crc : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Dial : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }

        class Execute : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Fgcolor : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }        
        class Getmatrix : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
       
        class Getprops : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Getptr : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Gradcolor : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }

        class Hammeraux : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Idct : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Inflate : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Interrupt : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Keys : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Loadidentity : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Loadimage : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Logo : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class March : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Memcpy : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Memcrc : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Memset : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Memwrite : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Memzero : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Number : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Progress : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Regread : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Rotate : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Scale : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Screensaver : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Scrollbar : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Setfont : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Setmatrix : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Sketch : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Slider : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Snapshot : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Spinner : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Stop : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }


       
        class Touch_transform : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Track : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
        class Translate : FTEveCmd
        {
            public CMD cmd()
            {
                throw new NotImplementedException();
            }

            public uint Size()
            {
                throw new NotImplementedException();
            }

            public byte[] ToBuffer()
            {
                throw new NotImplementedException();
            }
        }
#endif //

        public abstract  class FTEveCmd : IFTEveCmd
        {
            virtual public  byte [] CMDToBuffer ( CMD c )
            {

                byte[] buff = {     (byte)( ((uint)c & 0x000000FF)     )     ,
                                    (byte)( ((uint)c & 0x0000FF00) >> 8)     ,
                                    (byte)( ((uint)c & 0x00FF0000) >> 16)    ,
                                    (byte)( ((uint)c & 0xFF000000) >> 24 )   
                    };

                return buff;


            }

            public abstract uint Size();
            public abstract byte[] ToBuffer();
        }
    }
}
