﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Diagnostics;

using FTDI.D2XX.Device;
using Windows.UI;

// Notes:
//
// It is recommended to check all return values from the I2C functions, these have been omitted for clarity. 
// It is recommended to check if the I2C peripheral ACK'ed a write transaction by checking deviceAck after a write
// For example...
//
//  status = await SendI2CStart();
//  if (!status!)     Error
//  status = await SendByteAndCheckACK((byte)(0x52 | 0x00));
//  if (!status!)     Error
//  Check DeviceAck == True
//  status = await SendByteAndCheckACK((byte)(_ENABLE));      // "Enable" register
//  if (!status!)     Error
//  Check DeviceAck == True
//  status = await SendByteAndCheckACK((byte)(0x1B));      //      
//  if (!status!)     Error
//  Check DeviceAck == True
//  status = await SendI2CStop();
//  if (!status!)     Error
// ... where Error is the routine to handle the error condition.
//









namespace DeviceWatcher
{
    class Sensors
    {
        private FTManager ftManager;
        private FTD2xxDevice myDevice = null;
        

        // ####################################################################################################################################
        // ##################################################   DEFINES  ######################################################################
        // ####################################################################################################################################


        // ---------------------------------------- General Defines ----------------------------------------

        static UInt32 ftdiDeviceCount = 0;

        
        // ---------------------------------------- Defines for I2C Library ----------------------------------------
        
        public const byte I2C_Dir_SDAin_SCLin = 0x00;
        public const byte I2C_Dir_SDAin_SCLout = 0x01;
        public const byte I2C_Dir_SDAout_SCLout = 0x03;
        public const byte I2C_Dir_SDAout_SCLin = 0x02;
        public const byte I2C_Data_SDAhi_SCLhi = 0x03;
        public const byte I2C_Data_SDAlo_SCLhi = 0x01;
        public const byte I2C_Data_SDAlo_SCLlo = 0x00;

        public const uint ClockDivisor = 0x95;      // 200KHz
        
        public const byte MSB_FALLING_EDGE_CLOCK_BYTE_IN = 0x24;
        public const byte MSB_RISING_EDGE_CLOCK_BYTE_IN = 0x20;
        public const byte MSB_FALLING_EDGE_CLOCK_BYTE_OUT = 0x11;
        public const byte MSB_DOWN_EDGE_CLOCK_BIT_IN = 0x26;
        public const byte MSB_UP_EDGE_CLOCK_BYTE_IN = 0x20;
        public const byte MSB_UP_EDGE_CLOCK_BYTE_OUT = 0x10;
        public const byte MSB_RISING_EDGE_CLOCK_BIT_IN = 0x22;
        public const byte MSB_FALLING_EDGE_CLOCK_BIT_OUT = 0x13;
        
        static bool status = false;

        static byte[] MPSSEbuffer = new byte[500];
        static uint NumBytesToSend = 0;
        static uint NumBytesSent = 0;
        
        //static byte[] DataSend = new byte[500];         // 

        static byte[] InputBuffer = new byte[500];      // working buffer
        static byte[] InputBuffer2 = new byte[500];     // returning buffer
        static uint NumBytesToRead = 0;
        static uint BytesAvailable = 0;
        static uint NumBytesRead = 0;

        static bool DeviceAck = false;
        
        static byte GPIO_Low_Dat = 0;
        static byte GPIO_Low_Dir = 0;
        
        // ---------------------------------------- Defines for Color sensor ----------------------------------------

        static byte[] ColorData = new byte[500];
        static UInt16 GreenWord = 0;
        static UInt16 RedWord = 0;
        static UInt16 BlueWord = 0;     
        float Clear_Cal;
        float Red_Cal;
        float Green_Cal;
        float Blue_Cal;

        public const byte _ENABLE = 0x80;//Enablestatusandinterrupts
        public const byte _ATIME = 0x81;//RGBCADCtime
        public const byte _WTIME = 0x83;//Waittime
        public const byte _AILTL = 0x84;//RGBCInterruptlowtresholdlowbyte
        public const byte _AILTH = 0x85;//RGBCInterruptlowtresholdhighbyte
        public const byte _AIHTL = 0x86;//RGBCInterrupthightresholdlowbyte
        public const byte _AIHTH = 0x87;//RGBCInterrupthightresholdhighbyte
        public const byte _PERS = 0x8C;//Interruptpersistencefilters
        public const byte _CONFIG = 0x8D;//Configuration
        public const byte _CONTROL = 0x8F;//Gaincontrolregister
        public const byte _ID = 0x92;//DeviceID
        public const byte _STATUS = 0x93;//Devicestatus
        public const byte _CDATA = 0x94;//ClearADClowdataregister
        public const byte _CDATAH = 0x95;//ClearADChighdataregister
        public const byte _RDATA = 0x96;//REDADClowdataregister
        public const byte _RDATAH = 0x97;//REDADChighdataregister
        public const byte _GDATA = 0x98;//GREENADClowdataregister
        public const byte _GDATAH = 0x99;//GREENADChighdataregister
        public const byte _BDATA = 0x9A;//BLUEADClowdataregister
        public const byte _BDATAH = 0x9B;//BLUEADChighdataregister
        public const byte _COLOR_W_ADDRESS = 0x52;
        public const byte _COLOR_R_ADDRESS = 0x53;
        
        public const byte PURPLE_FLAG = 1;
        public const byte BLUE_FLAG = 2;
        public const byte CYAN_FLAG = 3;
        public const byte GREEN_FLAG = 4;
        public const byte PINK_FLAG = 5;
        public const byte RED_FLAG = 6;
        public const byte ORANGE_FLAG = 7;
        public const byte YELLOW_FLAG = 8;
        
        public const byte _GAIN_x4 = 0x01;
        public const byte _GAIN_x16 = 0x10;
        public const byte _GAIN_x60 = 0x11;

        static byte ClearColorSensorL = 0;
        static byte ClearColorSensorH = 0;
        static float ClearColorSensor = 0;

        static byte RedColorSensorL = 0;
        static byte RedColorSensorH = 0;
        static float RedColorSensor = 0;

        static byte GreenColorSensorL = 0;
        static byte GreenColorSensorH = 0;
        static float GreenColorSensor = 0;

        static byte BlueColorSensorL = 0;
        static byte BlueColorSensorH = 0;
        static float BlueColorSensor = 0;


        // ---------------------------------------- Defines for Proximity sensor ----------------------------------------

        static byte Command = 0x00;
        static byte prox_loop = 0x00;
        static UInt32 SummProxiValue = 0;
        static byte[] ProxData = new byte[500];
        static UInt16 ProxiValue = 0, AmbiValue = 0;
        static byte LoByte, HiByte;
        static UInt16 AverageProxiValue = 0;
        static byte InterruptStatus = 0;
        static bool ProxiFlag = false;
        static bool AmbiFlag = false;
        
        public const byte VCNL40x0_ADDRESS = 0x26; // 8bit address

        // registers
        public const byte REGISTER_COMMAND = 0x80;
        public const byte REGISTER_ID = 0x81;
        public const byte REGISTER_PROX_RATE = 0x82;
        public const byte REGISTER_PROX_CURRENT = 0x83;
        public const byte REGISTER_AMBI_PARAMETER = 0x84;
        public const byte REGISTER_AMBI_VALUE = 0x85;
        public const byte REGISTER_PROX_VALUE = 0x87;
        public const byte REGISTER_INTERRUPT_CONTROL = 0x89;
        public const byte REGISTER_INTERRUPT_LOW_THRES = 0x8a;
        public const byte REGISTER_INTERRUPT_HIGH_THRES = 0x8c;
        public const byte REGISTER_INTERRUPT_STATUS = 0x8e;
        public const byte REGISTER_PROX_TIMING = 0xf9;

        // Bits in Command register = 0x80
        public const byte COMMAND_ALL_DISABLE = 0x00;
        public const byte COMMAND_SELFTIMED_MODE_ENABLE = 0x01;
        public const byte COMMAND_PROX_ENABLE = 0x02;
        public const byte COMMAND_AMBI_ENABLE = 0x04;
        public const byte COMMAND_PROX_ON_DEMAND = 0x08;
        public const byte COMMAND_AMBI_ON_DEMAND = 0x10;
        public const byte COMMAND_MASK_PROX_DATA_READY = 0x20;
        public const byte COMMAND_MASK_AMBI_DATA_READY = 0x40;
        public const byte COMMAND_MASK_LOCK = 0x80;
        
        // Bits in Product ID Revision Register = 0x81;
        public const byte PRODUCT_MASK_REVISION_ID = 0x0f;
        public const byte PRODUCT_MASK_PRODUCT_ID = 0xf0;

        // Bits in Prox Measurement Rate register = 0x82;
        public const byte PROX_MEASUREMENT_RATE_2 = 0x00; // DEFAULT
        public const byte PROX_MEASUREMENT_RATE_4 = 0x01;
        public const byte PROX_MEASUREMENT_RATE_8 = 0x02;
        public const byte PROX_MEASUREMENT_RATE_16 = 0x03;
        public const byte PROX_MEASUREMENT_RATE_31 = 0x04;
        public const byte PROX_MEASUREMENT_RATE_62 = 0x05;
        public const byte PROX_MEASUREMENT_RATE_125 = 0x06;
        public const byte PROX_MEASUREMENT_RATE_250 = 0x07;
        public const byte PROX_MASK_MEASUREMENT_RATE = 0x07;

        // Bits in Proximity LED current setting = 0x83;
        public const byte PROX_MASK_LED_CURRENT = 0x3f; // DEFAULT = 2
        public const byte PROX_MASK_FUSE_PROG_ID = 0xc0;
        
        // Bits in Ambient Light Parameter register = 0x84;
        public const byte AMBI_PARA_AVERAGE_1 = 0x00;
        public const byte AMBI_PARA_AVERAGE_2 = 0x01;
        public const byte AMBI_PARA_AVERAGE_4 = 0x02;
        public const byte AMBI_PARA_AVERAGE_8 = 0x03;
        public const byte AMBI_PARA_AVERAGE_16 = 0x04;
        public const byte AMBI_PARA_AVERAGE_32 = 0x05; // DEFAULT
        public const byte AMBI_PARA_AVERAGE_64 = 0x06;
        public const byte AMBI_PARA_AVERAGE_128 = 0x07;
        public const byte AMBI_MASK_PARA_AVERAGE = 0x07;
        public const byte AMBI_PARA_AUTO_OFFSET_ENABLE = 0x08; // DEFAULT enable
        public const byte AMBI_MASK_PARA_AUTO_OFFSET = 0x08;
        public const byte AMBI_PARA_MEAS_RATE_1 = 0x00;
        public const byte AMBI_PARA_MEAS_RATE_2 = 0x10; // DEFAULT
        public const byte AMBI_PARA_MEAS_RATE_3 = 0x20;
        public const byte AMBI_PARA_MEAS_RATE_4 = 0x30;
        public const byte AMBI_PARA_MEAS_RATE_5 = 0x40;
        public const byte AMBI_PARA_MEAS_RATE_6 = 0x50;
        public const byte AMBI_PARA_MEAS_RATE_8 = 0x60;
        public const byte AMBI_PARA_MEAS_RATE_10 = 0x70;
        public const byte AMBI_MASK_PARA_MEAS_RATE = 0x70;
        public const byte AMBI_PARA_CONT_CONV_ENABLE = 0x80;
        public const byte AMBI_MASK_PARA_CONT_CONV = 0x80; // DEFAULT disable
        
        // Bits in Interrupt Control Register = x89;
        public const byte INTERRUPT_THRES_SEL_PROX = 0x00;
        public const byte INTERRUPT_THRES_SEL_ALS = 0x01;
        public const byte INTERRUPT_THRES_ENABLE = 0x02;
        public const byte INTERRUPT_ALS_READY_ENABLE = 0x04;
        public const byte INTERRUPT_PROX_READY_ENABLE = 0x08;
        public const byte INTERRUPT_COUNT_EXCEED_1 = 0x00; // DEFAULT
        public const byte INTERRUPT_COUNT_EXCEED_2 = 0x20;
        public const byte INTERRUPT_COUNT_EXCEED_4 = 0x40;
        public const byte INTERRUPT_COUNT_EXCEED_8 = 0x60;
        public const byte INTERRUPT_COUNT_EXCEED_16 = 0x80;
        public const byte INTERRUPT_COUNT_EXCEED_32 = 0xa0;
        public const byte INTERRUPT_COUNT_EXCEED_64 = 0xc0;
        public const byte INTERRUPT_COUNT_EXCEED_128 = 0xe0;
        public const byte INTERRUPT_MASK_COUNT_EXCEED = 0xe0;
        
        // Bits in Interrupt Status Register = x8e;
        public const byte INTERRUPT_STATUS_THRES_HI = 0x01;
        public const byte INTERRUPT_STATUS_THRES_LO = 0x02;
        public const byte INTERRUPT_STATUS_ALS_READY = 0x04;
        public const byte INTERRUPT_STATUS_PROX_READY = 0x08;
        public const byte INTERRUPT_MASK_STATUS_THRES_HI = 0x01;
        public const byte INTERRUPT_MASK_THRES_LO = 0x02;
        public const byte INTERRUPT_MASK_ALS_READY = 0x04;
        public const byte INTERRUPT_MASK_PROX_READY = 0x08;
        
        
        // ####################################################################################################################################
        // ################################################  END OF DEFINES ###################################################################
        // ####################################################################################################################################




        





        // ####################################################################################################################################
        // ####################################################################################################################################
        // #########################                               MAIN PROGRAM                                     ###########################
        // ####################################################################################################################################
        // ####################################################################################################################################

        
        public Sensors()
        {
            ftManager = new FTManager();
        }
        

        // ---------------------------------------- Opening the channel for the I2C sensors ----------------------------------------

        public async Task SensorChannelConfig()
        {
            // Determine the number of FTDI devices connected to the machine
                 
            //Create device list...
            var devList = ftManager.GetDeviceList();
            if (devList.Count == 0)
            {
                return;
            }

            string description = "";
            foreach (FTD2xxDevice d in FTManager.FtdiDeviceList)
            {
                if (d.DeviceInfoNode.Description.Contains("FT4232H Hub Module B"))
                {
                    description = d.DeviceInfoNode.Description;
                    myDevice = ftManager.OpenByDeviceID(d.DeviceInfoNode.DeviceId);
                    if (myDevice == null)
                    {
                        return;
                    }
                    await myDevice.OpenLite();
                    break;
                }
            }
            
            // Initialise the I2C interface 

            await InitMPSSE_For_I2C();

            // Configure the I2C sensors

            await ColourSensorConfig();       

            await ProximitySensorConfig();    


        } 

        public async Task<Proximity> GetProximity()
        {
            Proximity p;
            await ProximitySensorReading();       // See global variables ProxiFlag, ProviValue, AmbiFlag, AmbiValue
                                                  // Flags are true if the current reading has valid value, ignore associated reading if false
            p = new Proximity(ProxiValue, ProxiFlag, AmbiValue);
            return p;
        }

        public async Task<Color> GetColor()
        {
            Color c = await ColourSensorReadingAdafruit();
            return c;
        }




        // ####################################################################################################################################
        // ###########################################################  SENSOR API LAYER  #####################################################
        // ####################################################################################################################################

        public async Task<bool> ColourSensorConfig()
        {
            //Initialise Color sensor TCS3471

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(0x52 | 0x00));
            await SendByteAndCheckACK((byte)(_ENABLE));      // "Enable" register
            await SendByteAndCheckACK((byte)(0x1B));      //      
            await SendI2CStop();

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(0x52 | 0x00));
            await SendByteAndCheckACK((byte)(_CONTROL));      // 
            await SendByteAndCheckACK((byte)(_GAIN_x16));      //      
            await SendI2CStop();

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(0x52 | 0x00));
            await SendByteAndCheckACK((byte)(_ATIME));      // 
            await SendByteAndCheckACK((byte)(0x00));      //      
            await SendI2CStop();

            
            // Calculate an average reading which could be used to adjust the displayed results based on ambient levels
            // Provided for future extension of the application
            byte color_loop = 0;
            for (color_loop = 0; color_loop < 16; color_loop++)
            {

                await ColourSensorReadingAdafruit();

                Clear_Cal += ClearColorSensor;
                Red_Cal += RedColorSensor;
                Green_Cal += GreenColorSensor;
                Blue_Cal += BlueColorSensor;

                await Task.Delay(10);

            }

            Clear_Cal   = Clear_Cal / 16;
            Red_Cal     = Red_Cal / 16;
            Green_Cal   = Green_Cal / 16;
            Blue_Cal    = Blue_Cal / 16;

            return true;    
        }

        // ####################################################################################################################################

        public async Task<bool> ProximitySensorConfig()
        {

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));// 0 = write   address is 0x26 (or with 0x01 if it's read)
            await SendByteAndCheckACK((byte)(REGISTER_PROX_CURRENT));  // 
            await SendByteAndCheckACK((byte)(20));  // Value
            await SendI2CStop();

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));// 0 = write   address is 0x26 (or with 0x01 if it's read)
            await SendByteAndCheckACK((byte)(REGISTER_PROX_RATE));  //
            await SendByteAndCheckACK((byte)(PROX_MEASUREMENT_RATE_31));  // Value
            await SendI2CStop();

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));// 0 = write   address is 0x26 (or with 0x01 if it's read)
            await SendByteAndCheckACK((byte)(REGISTER_COMMAND));  // 
            await SendByteAndCheckACK((byte)(COMMAND_PROX_ENABLE | COMMAND_AMBI_ENABLE | COMMAND_SELFTIMED_MODE_ENABLE));  // Value
            await SendI2CStop();

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));// 0 = write   address is 0x26 (or with 0x01 if it's read)
            await SendByteAndCheckACK((byte)(REGISTER_INTERRUPT_CONTROL));  // 
            await SendByteAndCheckACK((byte)(INTERRUPT_THRES_SEL_PROX | INTERRUPT_THRES_ENABLE | INTERRUPT_COUNT_EXCEED_1));  // Value
            await SendI2CStop();

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));// 0 = write   address is 0x26 (or with 0x01 if it's read)
            await SendByteAndCheckACK((byte)(REGISTER_AMBI_PARAMETER));  // 
            await SendByteAndCheckACK((byte)(AMBI_PARA_AVERAGE_32 | AMBI_PARA_AUTO_OFFSET_ENABLE | AMBI_PARA_MEAS_RATE_2));  // Value
            await SendI2CStop();

            // Get an average of 16 proximity values which can be used to calibrate and set the 0 level of the displayed results.
            SummProxiValue = 0;
            for (prox_loop = 0; prox_loop < 16; prox_loop++)
            {
                do
                { // wait on prox data ready bit
                    await SendI2CStart();
                    await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));
                    await SendByteAndCheckACK((byte)(REGISTER_COMMAND));  // register "command"
                    await SendI2CStart();
                    await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x01));
                    await ReadByteAndSendNAK();
                    Command = InputBuffer2[0];
                    await SendI2CStop();
                } while ((Command & COMMAND_MASK_PROX_DATA_READY) == 0);

                await SendI2CStart();
                await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));
                await SendByteAndCheckACK((byte)(REGISTER_PROX_VALUE));
                await SendI2CStart();
                await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x01));
                await ReadByteAndSendACK();
                ProxData[0] = InputBuffer2[0];
                await ReadByteAndSendNAK();
                ProxData[1] = InputBuffer2[0];
                await SendI2CStop();
                ProxiValue = (UInt16)((ProxData[0] << 8) | ProxData[1]);
                SummProxiValue += ProxiValue; // Sum of all measured prox values
            }

            AverageProxiValue = (UInt16)(SummProxiValue / 16); // calculate average  

            LoByte = (byte)((AverageProxiValue + 20) & 0x00FF);
            HiByte = (byte)(((AverageProxiValue + 20) & 0xFF00) >> 8);

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));// 0 = write   address is 0x26 (or with 0x01 if it's read)
            await SendByteAndCheckACK((byte)(REGISTER_INTERRUPT_HIGH_THRES));  // 
            await SendByteAndCheckACK((byte)(HiByte));  // Value
            await SendI2CStop();

            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));// 0 = write   address is 0x26 (or with 0x01 if it's read)
            await SendByteAndCheckACK((byte)(REGISTER_INTERRUPT_HIGH_THRES + 1));  // 
            await SendByteAndCheckACK((byte)(LoByte));  // Value
            await SendI2CStop();


            GPIO_Low_Dat = 0x08;
            GPIO_Low_Dir = 0x08;

            return true;
        }

        // ####################################################################################################################################

        public async Task<Color> ColourSensorReadingAdafruit()
        {
            Color col = new Color();
            byte color_loop = 0;
            byte ClearColorSensorL = 0;
            byte ClearColorSensorH = 0;
            ushort ClearColorSensor = 0;

            byte RedColorSensorL = 0;
            byte RedColorSensorH = 0;
            ushort RedColorSensor = 0;

            byte GreenColorSensorL = 0;
            byte GreenColorSensorH = 0;
            ushort GreenColorSensor = 0;

            byte BlueColorSensorL = 0;
            byte BlueColorSensorH = 0;
            ushort BlueColorSensor = 0;

            // Read the eight bytes which correspond to the four 16-bit values of clear, red, green and blue. 
            await SendI2CStart();
            await SendByteAndCheckACK((byte)(0x52 | 0x00));     // Addess of the sensor for writing
            await SendByteAndCheckACK((byte)(0xB4));            // select address of CLEARL register
            await SendI2CStart();                               // repeat start
            await SendByteAndCheckACK((byte)(0x52 | 0x01));     // Addess of the sensor for reading

            await ReadByteAndSendACK();                         // Now read all 8 result bytes with a burst read
            ClearColorSensorL = InputBuffer2[0];
            await ReadByteAndSendACK();
            ClearColorSensorH = InputBuffer2[0];

            await ReadByteAndSendACK();
            RedColorSensorL = InputBuffer2[0];
            await ReadByteAndSendACK();
            RedColorSensorH = InputBuffer2[0];

            await ReadByteAndSendACK();
            GreenColorSensorL = InputBuffer2[0];
            await ReadByteAndSendACK();
            GreenColorSensorH = InputBuffer2[0];

            await ReadByteAndSendACK();
            BlueColorSensorL = InputBuffer2[0];
            await ReadByteAndSendNAK();                             // NAK the last one to tell the sensor we're finished reading
            BlueColorSensorH = InputBuffer2[0];

            await SendI2CStop();

            // Combine to make a 16-bit value per colour

            ClearColorSensor = (ushort)((ClearColorSensorH * 256) | ClearColorSensorL);
            RedColorSensor =   (ushort)((RedColorSensorH * 256) | RedColorSensorL);
            GreenColorSensor = (ushort)((GreenColorSensorH * 256) | GreenColorSensorL);
            BlueColorSensor =  (ushort)((BlueColorSensorH * 256) | BlueColorSensorL);

            // scale the values to give a close match between the object measured and the displayed RGB colour on the LCD screen
            if (ClearColorSensor > 0)   // if any light present
            {
                RedColorSensor      = (ushort)(RedColorSensor            / (0.002 *ClearColorSensor));
                GreenColorSensor    = (ushort)(( 0.768*GreenColorSensor) / (0.002 * ClearColorSensor));
                BlueColorSensor     = (ushort)(( 0.852*BlueColorSensor)  / (0.002 * ClearColorSensor));
            }
            else
            {
                RedColorSensor      = (ushort)(1.0 * (RedColorSensor >> 6));
                GreenColorSensor    = (ushort)(1.0 * (GreenColorSensor >> 6));
                BlueColorSensor     = (ushort)(1.0 * (BlueColorSensor >> 6));
            }


            col.R = (byte)(RedColorSensor <   255 ?   RedColorSensor : 0xFF);
            col.G = (byte)(GreenColorSensor < 255 ? GreenColorSensor : 0xFF);
            col.B = (byte)(BlueColorSensor <  255 ? BlueColorSensor  : 0xFF);

            return col;
        }

        // ####################################################################################################################################

        public async Task<bool> ProximitySensorReading()
        {

            ProxiFlag = false;
            AmbiFlag = false;

            // wait on data ready bit
            do
            {
                await SendI2CStart();
                await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));
                await SendByteAndCheckACK((byte)(REGISTER_COMMAND));  // register "command"
                await SendI2CStart();
                await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x01));
                await ReadByteAndSendNAK();
                Command = InputBuffer2[0];
                await SendI2CStop();
            } while ((Command & (COMMAND_MASK_PROX_DATA_READY )) == 0);

            // Once the result is ready, read the values
            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x00));
            await SendByteAndCheckACK((byte)(REGISTER_PROX_VALUE));
            await SendI2CStart();
            await SendByteAndCheckACK((byte)(VCNL40x0_ADDRESS | 0x01));
            await ReadByteAndSendACK();
            ProxData[0] = InputBuffer2[0];
            await ReadByteAndSendNAK();
            ProxData[1] = InputBuffer2[0];
            await SendI2CStop();
            ProxiValue = (UInt16)((ProxData[0] << 8) | ProxData[1]);
            ProxiValue -= (ProxiValue > AverageProxiValue) ? AverageProxiValue : (ushort)ProxiValue;
            ProxiFlag = true;
            
            // If proximity is > 1000 then turn on the white illumination LED on the colour sensor
            if(ProxiValue > 1000)
                GPIO_Low_Dat = 0x08;
            else
                GPIO_Low_Dat = 0x00;
            
            return true;
        }

        

        // ####################################################################################################################################
        // #################################################  END OF SENSOR API LAYER  ########################################################
        // ####################################################################################################################################










        // ####################################################################################################################################
        // ##############################################  I2C LOWER LAYER  ###################################################################
        // ####################################################################################################################################

        // ***** Sets the port up as MPSSE with settings for I2C *****

        // IN       GPIO_Low_Dat        (Global) Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dir        (Global) Byte containing the data directions for the GPIO on bits 3-8
        // IN       ClockDivisor        (Global) Contains the required clock divider to set the I2C clock rate
        // Return                       Returns True if MPSSE set up successfully

        // Note: the functions ReadByteAndSendACK and ReadByteAndSendNAK could be combined with a boolean value passed in to specify ACK or NAK

        private async Task<bool> InitMPSSE_For_I2C()
        {
            /***** Initial device configuration *****/

            await myDevice.SetTimeouts((ushort)5000, (ushort)5000);
            await myDevice.SetLatencyTimerAsync(16);
            await myDevice.SetFlowControlAsync((ushort)FLOW_CONTROL.RTS_CTS, 0x00, 0x00);
            //await myDevice.SetFlowControlAsync((ushort)FLOW_CONTROL.NONE, 0x00, 0x00);                 
            await myDevice.SetBitModeAsync((byte)BIT_MODE.RESET, (byte) 0x00);
            await myDevice.SetBitModeAsync((byte)BIT_MODE.MPSSE, 0x02);         // MPSSE mode        
            
            /***** Flush the buffer *****/
            await D2xx_FlushBuffer();
                        
            /***** Synchronize the MPSSE interface by sending bad command 0xAA *****/
            byte[] buf1 = new byte[1];

            NumBytesToSend = 0;
            buf1[NumBytesToSend++] = 0xAA;

            status = await MyD2xx_WriteBytes(buf1,NumBytesToSend);
            if (!status) return false; 
            
            NumBytesToRead = 2;
            status = await D2xx_ReadBytes();
            if (!status) return false;

            if ((InputBuffer2[0] == 0xFA) && (InputBuffer2[1] == 0xAA))
            {
                //Debug.WriteLine("Bad Command Echo successful");
            }
            else
            {
                return false;
            }

            /***** Synchronize the MPSSE interface by sending bad command 0xAB *****/
            NumBytesToSend = 0;
            buf1[NumBytesToSend++] = 0xAB;
            status = await  MyD2xx_WriteBytes(buf1,NumBytesToSend);
            if (!status) return false;

            NumBytesToRead = 2;
            status = await D2xx_ReadBytes();
            if (!status) return false;

            if ((InputBuffer2[0] == 0xFA) && (InputBuffer2[1] == 0xAB))
            {
                //Debug.WriteLine("Bad Command Echo successful");
            }
            else
            {
                return false;
            }

            byte[] MPSSEbuffer = new byte[10];

            NumBytesToSend = 0;
            MPSSEbuffer[NumBytesToSend++] = 0x8A; 	//Ensure disable clock divide by 5 for 60Mhz master clock
            MPSSEbuffer[NumBytesToSend++] = 0x97;	 //Ensure turn off adaptive clocking
            MPSSEbuffer[NumBytesToSend++] = 0x8C; 	//Enable 3 phase data clock, used by I2C to allow data on both clock edges
            // The SK clock frequency can be worked out by below algorithm with divide by 5 set as off
            // SK frequency  = 60MHz /((1 +  [(1 +0xValueH*256) OR 0xValueL])*2)
            MPSSEbuffer[NumBytesToSend++] = 0x86; 			//Command to set clock divisor
            MPSSEbuffer[NumBytesToSend++] = (byte)(ClockDivisor & 0x00FF);	//Set 0xValueL of clock divisor
            MPSSEbuffer[NumBytesToSend++] = (byte)((ClockDivisor >> 8) & 0x00FF);	//Set 0xValueH of clock divisor
            MPSSEbuffer[NumBytesToSend++] = 0x85; 			// loopback off
            MPSSEbuffer[NumBytesToSend++] = 0x80; 	//Command to set directions of lower 8 pins and force value on bits set as output 
            MPSSEbuffer[NumBytesToSend++] = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));  	// Set idle states on the I2C lines
            MPSSEbuffer[NumBytesToSend++] = (byte)(0x00 | I2C_Dir_SDAin_SCLin | (GPIO_Low_Dir & 0xF8));	// Set idle states on the I2C lines

            status = await MyD2xx_WriteBytes(MPSSEbuffer,NumBytesToSend);
            if (!status)
            {
                return false;
            }
            else
            {
                return true;
            }
        }





        // #################################################################################################################################### 

        // ***** Clocks in a byte and clocks out a NAK *****

        // IN       GPIO_Low_Dat        (Global) Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dir        (Global) Byte containing the data directions for the GPIO on bits 3-8
        // OUT      InputBuffer2[0]     (Global) Contains the received byte
        // Return                       Returns True if no errors sending the commands or reading back the ack value

        // Note: the functions ReadByteAndSendACK and ReadByteAndSendNAK could be combined with a boolean value passed in to specify ACK or NAK


        public async  Task<bool> ReadByteAndSendNAK()
        {

            byte ADbusVal = 0;
            byte ADbusDir = 0;

            byte[] MPSSEbuffer = new byte[13];

            NumBytesToSend = 0;

            // Clock one byte of data in
            ADbusVal = (byte) (0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte) (0x00 | I2C_Dir_SDAin_SCLout | (GPIO_Low_Dir & 0xF8));//make data input
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_RISING_EDGE_CLOCK_BYTE_IN;      // Clock data byte in
            MPSSEbuffer[NumBytesToSend++] = 0x00;
            MPSSEbuffer[NumBytesToSend++] = 0x00;                               // Data length of 0x0000 means 1 byte data to clock in

            // Now send the NAK by clocking out one bit...send NAK bit as '1'
            ADbusVal = (byte) (0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte) (0x00 | I2C_Dir_SDAout_SCLout | (GPIO_Low_Dir & 0xF8));//back to output
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_FALLING_EDGE_CLOCK_BIT_OUT;    // Clock data bit out
            MPSSEbuffer[NumBytesToSend++] = 0x00;                              // Length of 0 means 1 bit
            MPSSEbuffer[NumBytesToSend++] = 0xFF;                              // Data bit to send is a '1'

            // This command then tells the MPSSE to send any results gathered back immediately
            MPSSEbuffer[NumBytesToSend++] = 0x87;                                  //    ' Send answer back immediate command

            status = await MyD2xx_WriteBytes(MPSSEbuffer, NumBytesToSend);
            if (!status)
            {
                return false;
            }
            else
            {
                NumBytesToRead = 1;
                status = await D2xx_ReadBytes();
                if (!status)
                    return false;
                else
                    return true;

                // InputBuffer2[0] now contains the results
            }
        }


        // #################################################################################################################################### 

        // ***** Clocks in a byte and clocks out an ACK *****

        // IN       GPIO_Low_Dat        (Global) Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dir        (Global) Byte containing the data directions for the GPIO on bits 3-8
        // OUT      InputBuffer2[0]     (Global) Contains the received byte
        // Return                       Returns True if no errors sending the commands or reading back the ack value

        // Note: the functions ReadByteAndSendACK and ReadByteAndSendNAK could be combined with a boolean value passed in to specify ACK or NAK
        
        public async  Task<bool> ReadByteAndSendACK()
        {

            byte ADbusVal = 0;
            byte ADbusDir = 0;

            byte[] MPSSEbuffer = new byte[13];

            NumBytesToSend = 0;

            // Clock one byte of data in
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAin_SCLout | (GPIO_Low_Dir & 0xF8));//make data input
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_RISING_EDGE_CLOCK_BYTE_IN;      // Clock data byte in
            MPSSEbuffer[NumBytesToSend++] = 0x00;
            MPSSEbuffer[NumBytesToSend++] = 0x00;                               // Data length of 0x0000 means 1 byte data to clock in

            // Now send the ACK by clocking out one bit...send ACK bit as '0'
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAout_SCLout | (GPIO_Low_Dir & 0xF8));//back to output
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_FALLING_EDGE_CLOCK_BIT_OUT;    // Clock data bit out
            MPSSEbuffer[NumBytesToSend++] = 0x00;                              // Length of 0 means 1 bit
            MPSSEbuffer[NumBytesToSend++] = 0x00;                              // Data bit to send is a '1'

            // This command then tells the MPSSE to send any results gathered back immediately
            MPSSEbuffer[NumBytesToSend++] = 0x87;                                  //    ' Send answer back immediate command

            status = await MyD2xx_WriteBytes(MPSSEbuffer, NumBytesToSend);
            if (!status)
            {
                return false;
            }
            else
            {
                NumBytesToRead = 1;
                status = await D2xx_ReadBytes();
                if (!status)
                    return false;
                else
                    return true;

                // InputBuffer2[0] now contains the results
            }
        }



        // #################################################################################################################################### 

        // ***** Clocks in 2 bytes, ACKing the first and NAKing the second *****

        // IN       GPIO_Low_Dat        (Global) Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dir        (Global) Byte containing the data directions for the GPIO on bits 3-8
        // OUT      InputBuffer2[]      (Global) Contains the received bytes in locations [0] and [1]
        // Return                       Returns True if no errors sending the commands or reading back the ack value

        // Note: this function could be modified to receive other numbers of bytes if required, depending on the intended I2C peripheral
        
        public async  Task<bool> Read2BytesAndSendNAK()
        {

            byte ADbusVal = 0;
            byte ADbusDir = 0;

             byte[] MPSSEbuffer = new byte[25];

            NumBytesToSend = 0;

            // Clock one byte of data in
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAin_SCLout | (GPIO_Low_Dir & 0xF8));//make data input
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_RISING_EDGE_CLOCK_BYTE_IN;      // Clock data byte in
            MPSSEbuffer[NumBytesToSend++] = 0x00;
            MPSSEbuffer[NumBytesToSend++] = 0x00;                               // Data length of 0x0000 means 1 byte data to clock in

            // Now send the ACK by clocking out one bit...send ACK bit as '0'
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAout_SCLout | (GPIO_Low_Dir & 0xF8));//back to output
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_FALLING_EDGE_CLOCK_BIT_OUT;    // Clock data bit out
            MPSSEbuffer[NumBytesToSend++] = 0x00;                              // Length of 0 means 1 bit
            MPSSEbuffer[NumBytesToSend++] = 0x00;                              // Data bit to send is a '0'   ACK

            // ---------------
            
            // Clock one byte of data in
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAin_SCLout | (GPIO_Low_Dir & 0xF8));//make data input
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_RISING_EDGE_CLOCK_BYTE_IN;      // Clock data byte in
            MPSSEbuffer[NumBytesToSend++] = 0x00;
            MPSSEbuffer[NumBytesToSend++] = 0x00;                               // Data length of 0x0000 means 1 byte data to clock in

            // Now send the NAK by clocking out one bit...send NAK bit as '1'
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAout_SCLout | (GPIO_Low_Dir & 0xF8));//back to output
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_FALLING_EDGE_CLOCK_BIT_OUT;    // Clock data bit out
            MPSSEbuffer[NumBytesToSend++] = 0x00;                              // Length of 0 means 1 bit
            MPSSEbuffer[NumBytesToSend++] = 0xFF;                              // Data bit to send is a '1'    NAK
            

            // This command then tells the MPSSE to send any results gathered back immediately
            MPSSEbuffer[NumBytesToSend++] = 0x87;                                  //    ' Send answer back immediate command

            status = await MyD2xx_WriteBytes(MPSSEbuffer, NumBytesToSend);
            if (!status)
            {
                return false;
            }
            else
            {
                NumBytesToRead = 2;
                status = await D2xx_ReadBytes();
                if (!status)
                    return false;
                else
                    return true;

                // InputBuffer2[0] AND InputBuffer2[0] now contain the results
            }
        }

        // #################################################################################################################################### 

        // ***** Sends a byte and checks for an ACK *****
        
        // IN       DataByteToSend      Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dat        (Global) Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dir        (Global) Byte containing the data directions for the GPIO on bits 3-8
        // OUT      DeviceAck           (Global) Bool - True if the I2C peripheral ACK'ed the byte
        // Return                       Returns True if no errors sending the commands or reading back the ack value

        public async Task<bool> SendByteAndCheckACK(byte DataByteToSend)
        {
            
            byte ADbusVal = 0;
            byte ADbusDir = 0;

            byte[] MPSSEbuffer = new byte[16];
            
            NumBytesToSend = 0;

            // Clock one byte of data out
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAout_SCLout | (GPIO_Low_Dir & 0xF8));
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            MPSSEbuffer[NumBytesToSend++] = MSB_FALLING_EDGE_CLOCK_BYTE_OUT;          // ' clock data byte out
            MPSSEbuffer[NumBytesToSend++] = 0x00;                                  // '
            MPSSEbuffer[NumBytesToSend++] = 0x00;                                    //   ' Data length of 0x0000 means 1 byte data to clock in
            MPSSEbuffer[NumBytesToSend++] = DataByteToSend;                           //  ' Byte to send

            // Put line back to idle (data released, clock pulled low)
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAin_SCLout | (GPIO_Low_Dir & 0xF8));
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            // CLOCK IN ACK
            MPSSEbuffer[NumBytesToSend++] = MSB_RISING_EDGE_CLOCK_BIT_IN;           //   ' clock data byte in
            MPSSEbuffer[NumBytesToSend++] = 0x00;                                      // ' Length of 0 means 1 bit

            // SDA low, SCL low
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAout_SCLout | (GPIO_Low_Dir & 0xF8));
            MPSSEbuffer[NumBytesToSend++] = 0x80;                               //       ' Command - set low byte
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;                            //      ' Set the values
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;                             //     ' Set the directions

            // This command then tells the MPSSE to send any results gathered back immediately
            MPSSEbuffer[NumBytesToSend++] = 0x87;                                //  ' Send answer back immediate command

            status = await MyD2xx_WriteBytes(MPSSEbuffer, NumBytesToSend);
            if (!status)
            {
                return false;
            }
            else
            {
                NumBytesToRead = 1;
                status = await D2xx_ReadBytes();
                if (!status)
                {
                    return false;
                }
                else
                {
                    if (InputBuffer2[0] == 0)
                    {
                        DeviceAck = true;
                    }
                    else
                    {
                        DeviceAck = false;
                    }

                    return true;

                }
            }
        }



        // #################################################################################################################################### 

        // ***** Sends an I2C Start condition *****

        // IN       GPIO_Low_Dat        (Global) Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dir        (Global) Byte containing the data directions for the GPIO on bits 3-8
        // Return                       Returns True if the bytes were all sent

        public async Task<bool> SendI2CStart()
        {
            byte Count = 0;
            byte ADbusVal = 0;
            byte ADbusDir = 0;
            byte[] MPSSEbuffer = new byte[39];

            NumBytesToSend = 0;

            // Both SDA and SCL high
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAin_SCLin | (GPIO_Low_Dir & 0xF8));

            for (Count = 0; Count < 6; Count++)
            {
                MPSSEbuffer[NumBytesToSend++] = 0x80;	    // ADbus GPIO command
                MPSSEbuffer[NumBytesToSend++] = ADbusVal;   // Set data value
                MPSSEbuffer[NumBytesToSend++] = ADbusDir;	// Set direction
            }

            // SDA low, SCL high
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAout_SCLin | (GPIO_Low_Dir & 0xF8));

            for (Count = 0; Count < 6; Count++)	// Repeat commands to ensure the minimum period of the start setup time
            {
                MPSSEbuffer[NumBytesToSend++] = 0x80;	    // ADbus GPIO command
                MPSSEbuffer[NumBytesToSend++] = ADbusVal;   // Set data value
                MPSSEbuffer[NumBytesToSend++] = ADbusDir;	// Set direction
            }

            // SDA low, SCL low
            ADbusVal = (byte)(0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));    
            ADbusDir = (byte)(0x00 | I2C_Dir_SDAout_SCLout | (GPIO_Low_Dir & 0xF8));

            MPSSEbuffer[NumBytesToSend++] = 0x80;	    // ADbus GPIO command
            MPSSEbuffer[NumBytesToSend++] = ADbusVal;   // Set data value
            MPSSEbuffer[NumBytesToSend++] = ADbusDir;   // Set direction

            status = await MyD2xx_WriteBytes(MPSSEbuffer, NumBytesToSend);
            if (!status)
                return false;
            else
                return true;

            // At this point, both output and driving low
        }



        // ####################################################################################################################################

        // ***** Sends an I2C Stop condition *****

        // IN       GPIO_Low_Dat        (Global) Byte containing the data values for the GPIO on bits 3-8
        // IN       GPIO_Low_Dir        (Global) Byte containing the data directions for the GPIO on bits 3-8
        // Return                       Returns True if the bytes were all sent
        
        public async Task<bool> SendI2CStop()
        {
            byte Count = 0;
            byte ADbusVal = 0;
            byte ADbusDir = 0;

            byte[] MPSSEbuffer = new byte[36];
            
            NumBytesToSend = 0;

            // SDA low, SCL high, upper 5 bits defined by Global GPIO variables
            ADbusVal = (byte) (0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte) (0x00 | I2C_Dir_SDAout_SCLin | (GPIO_Low_Dir & 0xF8));

            for (Count = 0; Count < 6; Count++)
            {
                MPSSEbuffer[NumBytesToSend++] = 0x80;	    // ADbus GPIO command
                MPSSEbuffer[NumBytesToSend++] = ADbusVal;   // Set data value
                MPSSEbuffer[NumBytesToSend++] = ADbusDir;	// Set direction
            }

            // SDA high, SCL high, upper 5 bits defined by Global GPIO variables
            ADbusVal = (byte) (0x00 | I2C_Data_SDAlo_SCLlo | (GPIO_Low_Dat & 0xF8));
            ADbusDir = (byte) (0x00 | I2C_Dir_SDAin_SCLin | (GPIO_Low_Dir & 0xF8));

            for (Count = 0; Count < 6; Count++)	// Repeat commands to ensure the minimum period of the start setup time
            {
                MPSSEbuffer[NumBytesToSend++] = 0x80;	    // ADbus GPIO command
                MPSSEbuffer[NumBytesToSend++] = ADbusVal;   // Set data value
                MPSSEbuffer[NumBytesToSend++] = ADbusDir;	// Set direction
            }
            
            status = await MyD2xx_WriteBytes(MPSSEbuffer, NumBytesToSend);
            if (!status)
                return false;
            else
                return true;

            // At this point, both input and so pulled high
        }


        // ####################################################################################################################################
        // ##############################################  END OF I2C LOWER LAYER  ############################################################
        // ####################################################################################################################################






















        // ####################################################################################################################################
        // ########################################################  D2XX LAYER  ##############################################################
        // ####################################################################################################################################

        // ***** Flushes buffer *****

        public  async Task D2xx_FlushBuffer()
        {
            BytesAvailable = await myDevice.GetQueueStatus();	                        // Get the number of bytes in the receive buffer
            if ( (BytesAvailable > 0))
                NumBytesRead = await myDevice.ReadAsync(InputBuffer, BytesAvailable);  	// Read out the data from receive buffer and throw away
        }

        // ####################################################################################################################################

        // ***** Writes the specified number of bytes *****

        // IN       buf                 Containing the data to write
        // IN       bytesToWrite        Containing number of bytes to write
        // Return                       Returns True if the bytes were all sent

        public async Task<bool> MyD2xx_WriteBytes(byte[] buf, uint bytesToWrite)
        {
            bool retVal = false;
            
            NumBytesSent = await myDevice.WriteAsync(buf, bytesToWrite);
            if ((NumBytesSent == bytesToWrite))
            {
                retVal = true;
            }
            
            return retVal;
        }

        // ####################################################################################################################################

        // Reads bytes 
        // IN       NumBytesToRead      Global variable containing the number of bytes to read
        // OUT      NumBytesRead        Global variable containing actual number of bytes read
        // OUT      InputBuffer2[0]     Global buffer contains the bytes read
        // Return                       Returns True if all bytes requested were received

        public async Task<bool> D2xx_ReadBytes()
        {
            uint NumBytesInQueue = 0;
            uint QueueTimeOut = 0;
            uint Buffer1Index = 0;
            uint Buffer2Index = 0;
            uint TotalBytesRead = 0;
            bool QueueTimeoutFlag = false;
            uint NumBytesRxd = 0;

            // Keep looping until all requested bytes are received or we've tried 500 times
            while ((TotalBytesRead < NumBytesToRead) && (QueueTimeoutFlag == false))
            {
                NumBytesInQueue =  await myDevice.GetQueueStatus();

                if ((NumBytesInQueue > 0) )
                {
                    NumBytesRxd = await myDevice.ReadAsync (InputBuffer, NumBytesInQueue);  	//Read out the data from receive buffer

                    if ((NumBytesInQueue == NumBytesRxd) )
                    {
                        Buffer1Index = 0;

                        while (Buffer1Index < NumBytesRxd)
                        {
                            InputBuffer2[Buffer2Index] = InputBuffer[Buffer1Index];
                            Buffer1Index++;
                            Buffer2Index++;
                        }
                        TotalBytesRead = TotalBytesRead + NumBytesRxd;
                    }

                    QueueTimeOut++;
                    if (QueueTimeOut == 500)
                        QueueTimeoutFlag = true;
                    else
                        await Task.Delay(1);
                }
            }

            // returning NumBytesRead and the buffer InputBuffer2
            NumBytesRead = TotalBytesRead;

            if (QueueTimeoutFlag == true)
                return false;
            else
                return true;
        }



        // ####################################################################################################################################
        // ###############################################  END OF D2XX LAYER  ################################################################
        // ####################################################################################################################################

 

        public static void error()
        {
            ////////{
            ////////    Debug.WriteLine("Device initialisation failed");
            ////////    Console.ReadKey();
            ////////    return;
            ////////}
        }




        public ColourName GetColourName(Color inCol)

                {
                    ColourName col = ColourName.undefined;
                    bool color_detected = false;

                    double color_value = RGB_To_HSL(inCol.R, inCol.G, inCol.B);


                    Debug.WriteLine("h {0} R {1} G {2} B {3} ", color_value, inCol.R, inCol.G, inCol.B);

            if ((color_value >= 0.8) && (color_value <= 1.0000))
                col = ColourName.Red;
            else if ((color_value >= 0.690) && (color_value <= 0.8))
                col = ColourName.Purple;
            else if ((color_value >= 0.605) && (color_value <= 0.630))
                col = ColourName.Blue;
            else if ((color_value >= 0.54) && (color_value <= 0.605))
                col = ColourName.LtBlue;
            else if ((color_value >= 0.25) && (color_value <= 0.54))
                col = ColourName.Green;
            else if ((color_value >= 0.0800) && (color_value <= 0.25))
                col = ColourName.Orange;
            else if ((color_value >= 0.00) && (color_value < 0.0800))
                col = ColourName.Red;
    
            return col;
                }









        // Convert RGB values to HSL values
        private float RGB_To_HSL(float red, float green, float blue)
        {
            float fmax, fmin, hue, saturation, luminance;

            fmax = Math.Max(Math.Max(red, green), blue);
            fmin = Math.Min(Math.Min(red, green), blue);

            luminance = fmax;
            if (fmax > 0)
                saturation = (fmax - fmin) / fmax;
            else
                saturation = 0;

            if (saturation == 0)
                hue = 0;
            else 
            {
                if (fmax == red)
                    hue = (((green - blue) / (fmax - fmin) )%6);
                    //hue = ((green - blue) / (fmax - fmin)) ;
                else if (fmax == green)
                { 
                    hue = 2 + (blue - red) / (fmax - fmin);
                    hue = hue / 6;
                }
                else
                {
                    hue = 4 + (red - green) / (fmax - fmin);
                    hue = hue / 6;
                }

                if (hue < 0)
                    hue += 1;
            }
            return hue;
            
        }






    }


    // Names used to display the colour values
    public enum ColourName
    {
        Red,
        Orange,
        Yellow,
        Green,        
        Cyan,
        LtBlue,
        Blue,
        Purple,
        undefined
    }

  
    public class Proximity
    {
        private ushort ambient;
        private ushort prox;
        private bool valid;
       

        public ushort Ambient
        {
            get { return ambient; }
            set { ambient = value; }
        }
        public ushort Prox
        {
            get { return prox; }
            set { prox = value; }
        }

        public bool Valid
        {
            get { return valid; }
            set { valid = value; }
        }

        public Proximity(ushort inProx, bool inValid)
        {
            prox = inProx;
            valid = inValid;
            ambient = 0;
        }
                 
        public Proximity(ushort inProx, bool inValid, ushort inAmbient)
        {
            prox = inProx;
            valid = inValid;
            ambient = inAmbient;
        }
    }
}
